try:
    from . import generic as g
except BaseException:
    import generic as g


class SampleTest(g.unittest.TestCase):

    def test_sample(self):
        m = g.get_mesh('featuretype.STL')

        samples = m.sample(1000)

        # check to make sure all samples are on the mesh surface
        distance = m.nearest.signed_distance(samples)
        assert g.np.abs(distance).max() < 1e-4

        even, index = g.trimesh.sample.sample_surface_even(m, 1000)
        # check to make sure all samples are on the mesh surface
        distance = m.nearest.signed_distance(even)
        assert g.np.abs(distance).max() < 1e-4

    def test_weights(self):
        m = g.trimesh.creation.box()

        # weigh all faces except first as zero
        weights = g.np.zeros(len(m.faces))
        weights[0] = 1.0

        # sample with passed weights
        points, fid = m.sample(count=100,
                               return_index=True,
                               face_weight=weights)
        # all faces should be on single face
        assert (fid == 0).all()

        # oversample box to make sure weights aren't screwing
        # up ability to get every face when weighted by area
        assert set(g.np.unique(m.sample(
            100000, return_index=True)[1])) == set(range(len(m.faces)))


if __name__ == '__main__':
    g.trimesh.util.attach_to_log()
    g.unittest.main()



try:
    from . import generic as g
except BaseException:
    import generic as g


class IdentifierTest(g.unittest.TestCase):

    def test_identifier(self, count=25):
        meshes = g.np.append(list(g.get_meshes(
            only_watertight=True, split=True, min_volume=0.001)),
            g.get_mesh('fixed_top.ply'))
        for mesh in meshes:
            if not mesh.is_volume or mesh.body_count != 1:
                g.log.warning('Mesh %s is not watertight!',
                              mesh.metadata['file_name'])
                continue

            g.log.info('Trying hash at %d random transforms', count)
            hashed = []
            identifier = []
            for transform in g.random_transforms(count):
                permutated = mesh.copy().apply_transform(transform)
                hashed.append(permutated.identifier_hash)
                identifier.append(permutated.identifier)

            result = g.np.array(hashed)
            ok = (result[0] == result[1:]).all()

            if not ok:
                debug = [g.trimesh.util.sigfig_int(
                    i, g.trimesh.comparison.id_sigfig)
                    for i in identifier]
                g.log.error('Hashes on %s differ after transform:\n %s\n',
                            mesh.metadata['file_name'],
                            str(g.np.array(debug, dtype=g.np.int64)))
                raise ValueError('values differ after transform!')

            # stretch the mesh by a small amount
            stretched = mesh.copy().apply_scale(
                [0.99974507, 0.9995662, 1.0029832])
            if hashed[-1] == stretched.identifier_hash:
                raise ValueError(
                    'Hashes on %s didn\'t change after stretching',
                    mesh.metadata['file_name'])

    def test_scene_id(self):
        """
        A scene has a nicely constructed transform tree, so
        make sure transforming meshes around it doesn't change
        the nuts of their identifier hash.
        """
        scenes = [g.get_mesh('cycloidal.3DXML')]

        for s in scenes:
            for geom_name, mesh in s.geometry.items():
                meshes = []
                if not mesh.is_volume:
                    continue

                for node in s.graph.nodes_geometry:
                    T, geo = s.graph[node]
                    if geom_name != geo:
                        continue

                    m = mesh.copy()
                    m.apply_transform(T)
                    meshes.append(m)
                if not all(meshes[0].identifier_hash == i.identifier_hash
                           for i in meshes):
                    raise ValueError(
                        '{} differs after transform!'.format(geom_name))

        # check an example for a mirrored part
        assert (scenes[0].geometry['disc_cam_B'].identifier_hash !=
                scenes[0].geometry['disc_cam_A'].identifier_hash)

    def test_reflection(self):
        # identifier should detect mirroring
        a = g.get_mesh('featuretype.STL')
        b = a.copy()
        b.vertices[:, 2] *= -1.0
        b.invert()
        assert g.np.isclose(a.volume, b.volume)
        # hash should differ
        assert a.identifier_hash != b.identifier_hash

        # a mesh which is quite sensitive to mirroring
        a = g.get_mesh('mirror.ply')
        b = a.copy()
        b.vertices[:, 2] *= -1.0
        b.invert()
        assert g.np.isclose(a.volume, b.volume)
        # hash should differ
        assert a.identifier_hash != b.identifier_hash


if __name__ == '__main__':
    g.trimesh.util.attach_to_log()
    g.unittest.main()



try:
    from . import generic as g
except BaseException:
    import generic as g


class TrianglesTest(g.unittest.TestCase):

    def test_barycentric(self):
        for m in g.get_meshes(4):
            # a simple test which gets the barycentric coordinate at each of the three
            # vertices, checks to make sure the barycentric is [1,0,0] for the vertex
            # and then converts back to cartesian and makes sure the original points
            #  are the same as the conversion and back
            for method in ['cross', 'cramer']:
                for i in range(3):
                    barycentric = g.trimesh.triangles.points_to_barycentric(
                        m.triangles, m.triangles[:, i], method=method)
                    assert (g.np.abs(barycentric -
                                     g.np.roll([1.0, 0, 0], i)) < 1e-8).all()

                    points = g.trimesh.triangles.barycentric_to_points(
                        m.triangles, barycentric)
                    assert (g.np.abs(points - m.triangles[:, i]) < 1e-8).all()

    def test_closest(self):
        closest = g.trimesh.triangles.closest_point(
            triangles=g.data['triangles']['triangles'],
            points=g.data['triangles']['points'])

        comparison = (closest - g.data['triangles']['closest']).all()

        assert (comparison < 1e-8).all()
        g.log.info('finished closest check on %d triangles', len(closest))

    def test_closest_obtuse(self):
        # simple triangle in the xy-plane with an obtuse corner at vertex A
        ABC = g.np.float32([[0, 0, 0], [2, 0, 0], [-2, 1, 0]])
        D = g.np.float32([1, -1, 0])

        # ground truth: closest point from D is the center of the AB edge:
        # (1,0,0)
        gt_closest = g.np.float32([1, 0, 0])
        tm_closest = g.trimesh.triangles.closest_point([ABC], [D])[0]

        assert g.np.linalg.norm(gt_closest - tm_closest) < g.tol.merge

        # create a circle of points around the triangle
        # with a radius so that all points are outside of the triangle
        radius = 3
        nPtsOnCircle = 100
        alphas = g.np.linspace(
            g.np.pi / nPtsOnCircle,
            g.np.pi * 2 - g.np.pi / nPtsOnCircle,
            nPtsOnCircle)
        ptsOnCircle = g.np.transpose(
            [g.np.cos(alphas), g.np.sin(alphas), g.np.zeros(nPtsOnCircle)]) * radius

        def norm(v):
            return g.np.sqrt(g.np.einsum('...i,...i', v, v))

        def distToLine(o, v, p):
            return norm((o - p) - g.np.dot(o - p, v) * v)

        def distPointToEdge(U, V, P):  # edge [U, V], point P
            UtoV = V - U
            UtoP = P - U
            VtoP = P - V

            if g.np.dot(UtoV, UtoP) <= 0:
                # P is 'behind' U
                return norm(UtoP)
            elif g.np.dot(-UtoV, VtoP) <= 0:
                # P is 'behind' V
                return norm(VtoP)
            else:
                # P is 'between' U and V
                return distToLine(U, UtoV / norm(UtoV), P)

        # get closest points from trimesh and compute distances to the circle
        # points
        tm_dists = norm(
            ptsOnCircle -
            g.trimesh.triangles.closest_point(
                [ABC] *
                nPtsOnCircle,
                ptsOnCircle))
        # compute naive point-to-edge distances for all points and take the min of
        # the three edges
        gt_dists = g.np.float32([[distPointToEdge(ABC[i], ABC[(i + 1) % 3], pt)
                                  for i in range(3)] for pt in ptsOnCircle]).min(axis=1)
        diff_dists = tm_dists - gt_dists

        assert g.np.dot(diff_dists, diff_dists) < g.tol.merge

    def test_degenerate(self):
        tri = [[[0, 0, 0],
                [1, 0, 0],
                [-.5, 0, 0]],
               [[0, 0, 0],
                [0, 0, 0],
                [10, 10, 0]],
               [[0, 0, 0],
                [0, 0, 2],
                [0, 0, 2.2]],
               [[0, 0, 0],
                [1, 0, 0],
                [0, 1, 0]]]

        tri_gt = [False,
                  False,
                  False,
                  True]

        r = g.trimesh.triangles.nondegenerate(tri)
        assert len(r) == len(tri)
        assert (r == tri_gt).all()

    def test_angles(self):
        # a zero- area triangle
        tris = g.np.array(
            [[[0, 0, 0], [1, 0, 0], [1, 0, 0]]], dtype=g.np.float64)
        angles = g.trimesh.triangles.angles(tris)
        # degenerate angles should be zero, not NaN
        print(angles)
        assert g.np.allclose(angles, 0.0)

        # an equilateral triangle
        tris = g.np.array(
            [[[-1, 0, 0],
              [1, 0, 0],
              [0, g.np.sqrt(3), 0]]], dtype=g.np.float64)
        angles = g.trimesh.triangles.angles(tris)
        # degenerate angles should be zero, not NaN
        assert g.np.allclose(angles, g.np.radians(60))

        # an equilateral triangle transformed into space
        tris = g.trimesh.transform_points(
            g.np.array(
                [[-1, 0, 0],
                 [1, 0, 0],
                 [0, g.np.sqrt(3), 0]], dtype=g.np.float64),
            g.trimesh.transformations.random_rotation_matrix()).reshape((-1, 3, 3))
        angles = g.trimesh.triangles.angles(tris)
        # all angles should be 60 degrees
        assert g.np.allclose(angles, g.np.radians(60))

        # an 3-4-5 right triangle
        tris = g.trimesh.transform_points(
            g.np.array(
                [[0, 0, 0],
                 [3, 0, 0],
                 [0, 4, 0]], dtype=g.np.float64),
            g.trimesh.transformations.random_rotation_matrix()).reshape((-1, 3, 3))
        # get angles
        angles = g.trimesh.triangles.angles(tris)
        # make sure they match a 3-4-5
        assert g.np.allclose(
            g.np.sort(angles.ravel()),
            [g.np.arcsin(3.0 / 5), g.np.arcsin(4.0 / 5), g.np.pi / 2])


if __name__ == '__main__':
    g.trimesh.util.attach_to_log()
    g.unittest.main()



try:
    from . import generic as g
except BaseException:
    import generic as g


class OBJTest(g.unittest.TestCase):

    def test_rabbit(self):
        # A BSD-licensed test model from pyglet
        # it has mixed triangles, quads, and 16 element faces -_-
        # this should test the non-vectorized load path
        m = g.get_mesh('rabbit.obj')
        assert len(m.faces) == 1252
        rec = g.wrapload(
            m.export(file_type='obj'), file_type='obj')
        assert g.np.isclose(m.area, rec.area)

    def test_no_img(self):
        # sometimes people use the `vt` parameter for arbitrary
        # vertex attributes and thus want UV coordinates even
        # if there is no texture image
        m = g.get_mesh('noimg.obj')
        assert m.visual.uv.shape == (len(m.vertices), 2)
        # make sure UV coordinates are in range 0.0 - 1.0
        assert m.visual.uv.max() < (1 + 1e-5)
        assert m.visual.uv.min() > -1e-5
        # check to make sure it's not all zeros
        assert m.visual.uv.ptp() > 0.5
        rec = g.wrapload(
            m.export(file_type='obj'), file_type='obj')
        assert g.np.isclose(m.area, rec.area)

    def test_trailing(self):
        # test files with texture and trailing slashes
        m = g.get_mesh('jacked.obj')
        assert len(m.visual.uv) == len(m.vertices)
        rec = g.wrapload(
            m.export(file_type='obj'), file_type='obj')
        assert g.np.isclose(m.area, rec.area)

    def test_obj_groups(self):
        # a wavefront file with groups defined
        mesh = g.get_mesh('groups.obj')

        # make sure some data got loaded
        assert g.trimesh.util.is_shape(mesh.faces, (-1, 3))
        assert g.trimesh.util.is_shape(mesh.vertices, (-1, 3))

        # make sure groups are the right length
        # TODO: we do not support face groups now
        # assert len(mesh.metadata['face_groups']) == len(mesh.faces)

        # check to make sure there is signal not just zeros
        # assert mesh.metadata['face_groups'].ptp() > 0

    def test_obj_negative_indices(self):
        # a wavefront file with negative indices
        mesh = g.get_mesh('negative_indices.obj')

        # make sure some data got loaded
        assert g.trimesh.util.is_shape(mesh.faces, (12, 3))
        assert g.trimesh.util.is_shape(mesh.vertices, (8, 3))

    def test_obj_quad(self):
        mesh = g.get_mesh('quadknot.obj')
        # make sure some data got loaded
        assert g.trimesh.util.is_shape(mesh.faces, (-1, 3))
        assert g.trimesh.util.is_shape(mesh.vertices, (-1, 3))

        assert mesh.is_watertight
        assert mesh.is_winding_consistent
        rec = g.wrapload(
            mesh.export(file_type='obj'), file_type='obj')
        assert g.np.isclose(mesh.area, rec.area)

    def test_obj_multiobj(self):
        # test a wavefront file with multiple objects in the same file
        scene = g.get_mesh('two_objects.obj',
                           split_object=True,
                           group_material=False)
        assert len(scene.geometry) == 2

        for mesh in scene.geometry.values():
            # make sure some data got loaded
            assert g.trimesh.util.is_shape(mesh.faces, (-1, 3))
            assert g.trimesh.util.is_shape(mesh.vertices, (-1, 3))

            assert mesh.is_watertight
            assert mesh.is_winding_consistent

    def test_obj_split_attributes(self):
        # test a wavefront file where pos/uv/norm have different indices
        # and where multiple objects share vertices
        # Note 'process=False' to avoid merging vertices
        scene = g.get_mesh('joined_tetrahedra.obj',
                           process=False,
                           split_object=True,
                           group_material=False)

        assert len(scene.geometry) == 2

        geom = list(scene.geometry.values())

        assert g.trimesh.util.is_shape(geom[0].faces, (4, 3))
        assert g.trimesh.util.is_shape(geom[0].vertices, (9, 3))
        assert g.trimesh.util.is_shape(geom[1].faces, (4, 3))
        assert g.trimesh.util.is_shape(geom[1].vertices, (9, 3))

    def test_obj_simple_order(self):

        # test a simple wavefront model without split indexes
        # and make sure we don't reorder vertices unnecessarily
        file_name = g.os.path.join(g.dir_models, 'cube.OBJ')
        # load a simple OBJ file without merging vertices
        m = g.trimesh.load(file_name, process=False)
        # use trivial loading to compare with fancy performant one
        with open(file_name, 'r') as f:
            f, v, vt = simple_load(f.read())
        # trimesh loader should return the same face order
        assert g.np.allclose(f, m.faces)
        assert g.np.allclose(v, m.vertices)

    def test_order_tex(self):

        # test a simple wavefront model without split indexes
        # and make sure we don't reorder vertices unnecessarily
        file_name = g.os.path.join(g.dir_models, 'fuze.obj')
        # load a simple OBJ file without merging vertices
        m = g.trimesh.load(
            file_name,
            process=False,
            maintain_order=True)
        # use trivial loading to compare with fancy performant one
        with open(file_name, 'r') as f:
            f, v, vt = simple_load(f.read())
        # trimesh loader should return the same face order
        assert g.np.allclose(f, m.faces)
        assert g.np.allclose(v, m.vertices)

    def test_obj_compressed(self):
        mesh = g.get_mesh('cube_compressed.obj', process=False)
        assert mesh._cache.cache['vertex_normals'].shape == mesh.vertices.shape
        assert g.np.allclose(
            g.np.abs(mesh.vertex_normals).sum(axis=1), 1.0)

    def test_vertex_color(self):
        # get a box mesh
        mesh = g.trimesh.creation.box()
        # set each vertex to a unique random color
        mesh.visual.vertex_colors = [
            g.trimesh.visual.random_color()
            for _ in range(len(mesh.vertices))]
        # export and then reload the file as OBJ
        rec = g.trimesh.load(
            g.trimesh.util.wrap_as_stream(
                mesh.export(file_type='obj')),
            file_type='obj')
        # assert colors have survived the export cycle
        assert (mesh.visual.vertex_colors ==
                rec.visual.vertex_colors).all()

    def test_single_vn(self):
        """
        Make sure files with a single VN load.
        """
        m = g.get_mesh('singlevn.obj')
        assert len(m.vertices) > 0
        assert len(m.faces) > 0

    def test_polygon_faces(self):
        m = g.get_mesh('polygonfaces.obj')
        assert len(m.vertices) > 0
        assert len(m.faces) > 0

    def test_faces_not_enough_indices(self):
        m = g.get_mesh('notenoughindices.obj')
        assert len(m.vertices) > 0
        assert len(m.faces) == 1

    def test_export_path(self):
        m = g.get_mesh('fuze.obj')
        g.check_fuze(m)
        assert m._cache.cache['vertex_normals'].shape == m.vertices.shape
        with g.TemporaryDirectory() as d:
            file_path = g.os.path.join(d, 'fz.obj')
            m.export(file_path)
            r = g.trimesh.load(file_path)
            g.check_fuze(r)

    def test_mtl(self):
        # get a mesh with texture
        m = g.get_mesh('fuze.obj')
        # export the mesh including data
        obj, data = g.trimesh.exchange.export.export_obj(
            m, return_texture=True)
        with g.trimesh.util.TemporaryDirectory() as path:
            # where is the OBJ file going to be saved
            obj_path = g.os.path.join(path, 'test.obj')
            with open(obj_path, 'w') as f:
                f.write(obj)
            # save the MTL and images
            for k, v in data.items():
                with open(g.os.path.join(path, k), 'wb') as f:
                    f.write(v)
            # reload the mesh from the export
            rec = g.trimesh.load(obj_path)
        # make sure loaded image is the same size as the original
        assert (rec.visual.material.image.size ==
                m.visual.material.image.size)
        # make sure the faces are the same size
        assert rec.faces.shape == m.faces.shape

    def test_scene(self):
        s = g.get_mesh('cycloidal.3DXML')

        e = g.trimesh.load(
            g.io_wrap(s.export(file_type='obj')),
            file_type='obj',
            split_object=True,
            group_materials=False)

        assert g.np.isclose(e.area, s.area, rtol=.01)

    def test_edge_cases(self):
        # a mesh with some NaN colors
        n = g.get_mesh('nancolor.obj')
        assert n.faces.shape == (12, 3)

        v = g.get_mesh('cubevt.obj')
        assert v.faces.shape == (12, 3)

    def test_empty_or_pointcloud(self):
        # demo files to check
        empty_files = ['obj_empty.obj',
                       'obj_points.obj',
                       'obj_wireframe.obj']

        for empty_file in empty_files:
            e = g.get_mesh('emptyIO/' + empty_file)

            # create export
            if 'empty' in empty_file:
                try:
                    export = e.export(file_type='ply')
                except BaseException:
                    continue
                raise ValueError('cannot export empty')
            elif 'points' in empty_file:
                export = e.export(file_type='ply')
                reconstructed = g.wrapload(export, file_type='ply')

                # result should be a point cloud instance
                assert isinstance(e, g.trimesh.PointCloud)
                assert hasattr(e, 'vertices')
                # point cloud export should contain vertices
                assert isinstance(reconstructed, g.trimesh.PointCloud)
                assert hasattr(reconstructed, 'vertices')

    def test_backslash_continuation_character(self):
        # an obj file with \ (backslash) line continuation characters
        m = g.get_mesh('wallhole.obj')
        assert m.faces.shape == (66, 3)

    def test_no_uv(self):
        mesh = g.get_mesh('box.obj')
        rec = g.wrapload(
            mesh.export(file_type='obj'), file_type='obj')
        assert g.np.isclose(mesh.area, rec.area)

    def test_no_uv_but_mtl(self):
        sphere = g.trimesh.creation.uv_sphere()
        sphere.visual = g.trimesh.visual.TextureVisuals(
            uv=None,
            material=g.trimesh.visual.material.empty_material())
        output = sphere.export('sphere.obj')
        assert 'usemtl' in output

    def test_chair(self):
        mesh = next(iter(g.get_mesh('chair.zip').geometry.values()))

        # this model comes with vertex normals
        assert 'vertex_normals' in mesh._cache
        assert g.np.allclose(
            1.0, g.np.linalg.norm(mesh.vertex_normals, axis=1))
        mesh.apply_scale(0.46377314288075433)
        assert 'vertex_normals' in mesh._cache
        assert g.np.allclose(
            1.0, g.np.linalg.norm(mesh.vertex_normals, axis=1))
        assert 'vertex_normals' in mesh._cache
        mesh._cache.clear()
        assert 'vertex_normals' not in mesh._cache
        # if we recomputed now, the degenerate faces
        # would lead some of these vertex normals to be zero
        # assert g.np.allclose(
        #    1.0, g.np.linalg.norm(mesh.vertex_normals, axis=1))

    def test_multi_nodupe(self):
        s = g.get_mesh("forearm.zae")
        obj, mtl = g.trimesh.exchange.obj.export_obj(
            s, include_color=True,
            include_texture=True,
            return_texture=True)
        # should be using one material file
        assert obj.count('mtllib') == 1
        assert 'mtllib material.mtl' in obj
        # should be specifying 5 materials
        assert obj.count('usemtl') == 5

        # this file has only the properties (no images)
        assert len(mtl) == 1
        mtl_names = [
            L.strip().split()[-1].strip() for L in
            mtl['material.mtl'].decode('utf-8').split('\n')
            if 'newmtl' in L]
        # there should be 5 unique material names
        assert len(set(mtl_names)) == 5

    def test_mtl_color_roundtrip(self):

        # create a mesh with a simple material
        m = g.trimesh.creation.box()
        m.visual = m.visual.to_texture()
        # set each color component to a unique value
        colors = [g.trimesh.visual.color.random_color()
                  for _ in range(3)]
        m.visual.material.ambient = colors[0]
        m.visual.material.specular = colors[1]
        m.visual.material.diffuse = colors[2]
        m.visual.material.glossiness = 0.52622

        with g.trimesh.util.TemporaryDirectory() as d:
            # exporting by filename will automatically
            # create a FilePathResolver which writes the
            # `mtl` file to the same directory
            file_name = g.os.path.join(d, 'hi.obj')
            m.export(file_name)
            # reload the export by file name
            r = g.trimesh.load(file_name)

        # these values should have survived the roundtrip
        assert g.np.allclose(m.visual.material.ambient,
                             r.visual.material.ambient)
        assert g.np.allclose(m.visual.material.specular,
                             r.visual.material.specular)
        assert g.np.allclose(m.visual.material.diffuse,
                             r.visual.material.diffuse)
        assert g.np.isclose(m.visual.material.glossiness,
                            r.visual.material.glossiness)

    def test_scene_export_material_name(self):
        s = g.get_mesh('fuze.obj', force='scene')
        dummy = 'fuxx'
        s.geometry['fuze.obj'].visual.material.name = dummy

        r = g.trimesh.resolvers.ZipResolver()
        r['model.obj'] = s.export(
            file_type='obj',
            mtl_name='mystuff.mtl',
            resolver=r)

        mtl = r['mystuff.mtl'].decode('utf-8')
        assert mtl.count('newmtl') == 1
        assert 'newmtl {}'.format(dummy) in mtl
        assert '{}.jpeg'.format(dummy) in r

    def test_compound_scene_export(self):

        # generate a mesh with multiple textures
        a = g.get_mesh('BoxTextured.glb')
        a = a.scaled(1.0 / a.extents.max())
        a.apply_translation(-a.bounds[0])

        b = g.get_mesh('fuze.obj').scene()
        b = b.scaled(1.0 / b.extents.max())
        b.apply_translation(-b.bounds[0] + [2, 0, 0])

        d = next(iter(b.copy().geometry.values()))
        d.apply_translation([-1, 0, 0])
        assert hash(d.visual.material) == hash(
            b.geometry['fuze.obj'].visual.material)

        # should change the material hash
        d.visual.material.glossiness = 0.1
        assert hash(d.visual.material) != hash(
            b.geometry['fuze.obj'].visual.material)

        # generate a compound scene
        c = a + b + d
        for i in c.geometry.values():
            # name all the materials the same thing
            i.visual.material.name = 'material_0'

        # export the compound scene
        obj, mtl = c.export(file_type='obj', return_texture=True)
        # there should be exactly one mtllib referenced
        assert obj.count('mtllib') == 1
        assert obj.count('usemtl') == 3

        # should be one texture image for each of 3
        # plus the `.mtl` file itself
        # if we had image-hash-deduplication this should
        # be changed to 3 as the image for `b` and `d` are the same
        assert len(mtl) == 4

        # get the material names specified

        mtl_names = [
            L.strip().split()[-1].strip() for L in
            mtl['material.mtl'].decode('utf-8').split('\n')
            if 'newmtl' in L]
        # there should be 3 unique material names
        assert len(set(mtl_names)) == 3

        # now reload the compound scene
        t = g.trimesh.load(
            file_obj=g.trimesh.util.wrap_as_stream(obj),
            file_type='obj',
            resolver=g.trimesh.resolvers.ZipResolver(mtl),
            group_material=False,
            split_object=True)
        # these names should match eventually
        assert len(t.geometry.keys()) == len(c.geometry.keys())
        assert g.np.isclose(t.area, c.area)

    def test_face_parsing_in_group_names(self):
        # Checks that an obj with a g tag containinig a face like name (an 'f '
        # followed by three space separated text chunks, ex: f 1 2 3) does load
        # properly
        m = g.get_mesh('face_in_group_name.obj')
        assert len(m.vertices) == 1


def simple_load(text):
    # we're going to load faces in a basic text way
    # and compare the order from this method to the
    # trimesh loader, to see if we get the same thing
    # note that trimesh's extremely convoluted string
    # wangling is wildly faster than this
    f = []
    v = []
    vt = []
    for line in str.splitlines(text):
        line = line.strip()
        if len(line) < 2:
            continue
        elif line.startswith('f '):
            if '/' in line:
                f.append([int(i.split('/', 1)[0])
                          for i in line[1:].strip().split()])
            else:
                f.append(line[1:].strip().split())
        elif line.startswith('v '):
            v.append(line[1:].strip().split())
        elif line.startswith('vt '):
            vt.append(line[2:].strip().split())

    # get faces as basic numpy array
    f = g.np.array(f, dtype=g.np.int64) - 1
    v = g.np.array(v, dtype=g.np.float64)
    vt = g.np.array(vt, dtype=g.np.float64)

    return f, v, vt


if __name__ == '__main__':
    g.trimesh.util.attach_to_log()
    g.unittest.main()



try:
    from . import generic as g
except BaseException:
    import generic as g


class TextureTest(g.unittest.TestCase):

    def test_uv_to_color(self):
        try:
            import PIL.Image
        except ImportError:
            return

        # n_vertices = 100
        uv = g.np.array([[0.25, 0.2], [0.4, 0.5]], dtype=float)
        texture = g.np.arange(96, dtype=g.np.uint8).reshape(8, 4, 3)
        colors = g.trimesh.visual.uv_to_color(
            uv, PIL.Image.fromarray(texture))

        colors_expected = [[75, 76, 77, 255], [51, 52, 53, 255]]

        assert (colors == colors_expected).all()

    def test_bad_uv(self):
        # get a textured OBJ
        m = g.get_mesh('fuze.obj', force='mesh')
        # add malformed UV coordinates
        m.visual.uv = m.visual.uv[:100]
        m.merge_vertices()

    def test_order_kwarg(self):
        for file_name in ['ico4.obj', 'ico4uv.obj']:
            # get the location of the model file
            file_path = g.get_path(file_name)
            with open(file_path, 'r') as f:
                # get the raw ordered vertices from the file with basic string
                # ops
                v_raw = g.np.array(
                    [line[2:].split() for line in f if line.startswith('v ')],
                    dtype=g.np.float64)

            # load them with maximal correspondence captain
            a = g.trimesh.load(file_path, process=False, maintain_order=True)
            # see if we have the same vertices
            assert g.np.allclose(a.vertices, v_raw)

            # load them without process and but without maintaining order
            a = g.trimesh.load(file_path, process=False, maintain_order=False)
            # vertex shape should not be the same
            assert a.vertices.shape != v_raw.shape

    def test_fuze(self):

        # create a local web server to test remote assets
        with g.serve_meshes() as address:
            # see if web resolvers work
            tex = g.trimesh.exchange.load.load_remote(
                url=address + '/fuze.obj', process=False)
            g.check_fuze(tex)

            # see if web + zip resolvers work
            scene = g.trimesh.exchange.load.load_remote(
                url=address + '/fuze.zip', process=False)

            # zip files get loaded into a scene
            assert len(scene.geometry) == 1
            # scene should just be a fuze bottle
            g.check_fuze(next(iter(scene.geometry.values())))

        # obj with texture, assets should be loaded
        # through a FilePathResolver
        m = g.get_mesh('fuze.obj', process=False)
        g.check_fuze(tex)

        # obj with texture, assets should be loaded
        # through a ZipResolver into a scene
        scene = g.get_mesh('fuze.zip', process=False)

        # zip files get loaded into a scene
        assert len(scene.geometry) == 1
        m = next(iter(scene.geometry.values()))
        g.check_fuze(m)

        # the PLY should have textures defined
        m = g.get_mesh('fuze.ply', process=False)
        g.check_fuze(m)

        # ASCII PLY should have textures defined
        m = g.get_mesh('fuze_ascii.ply', process=False)
        g.check_fuze(m)

        # textured meshes should subdivide OK-ish
        s = m.subdivide()
        assert len(s.visual.uv) == len(s.vertices)

        # load without doing the vertex separation
        # will look like garbage but represents original
        # and skips "disconnect vertices with different UV"
        b = g.get_mesh('fuze.ply',
                       process=False,
                       fix_texture=False)
        assert len(b.vertices) == 502
        assert len(b.visual.uv) == 502

    def test_upsize(self):
        """
        Texture images usually want to have sizes that are powers
        of two so resize textures up to the nearest power of two.
        """
        try:
            from PIL import Image
        except BaseException:
            g.log.warning('no PIL, not testing power_resize!')
            return

        # shortcut for the function
        resize = g.trimesh.visual.texture.power_resize

        img = Image.new('RGB', (10, 20))
        assert img.size == (10, 20)
        assert resize(img).size == (16, 32)
        assert resize(img, square=True).size == (32, 32)

        # check with one value on-size
        img = Image.new('RGB', (10, 32))
        assert img.size == (10, 32)
        assert resize(img).size == (16, 32)
        assert resize(img, square=True).size == (32, 32)

        # check early exit pathOA
        img = Image.new('RGB', (32, 32))
        assert img.size == (32, 32)
        assert resize(img).size == (32, 32)
        assert resize(img, square=True).size == (32, 32)

    def test_concatenate(self):
        # test concatenation with texture
        a = g.get_mesh('fuze.obj')
        b = a.copy()
        b.apply_translation([b.extents[0] * 1.25, 0, 0])

        c = a + b
        assert len(c.vertices) > len(a.vertices)
        assert len(c.visual.uv) == len(c.vertices)
        # should have deduplicated image texture
        assert g.np.allclose(c.visual.material.image.size,
                             a.visual.material.image.size)

    def test_to_tex(self):
        m = g.trimesh.creation.box()
        color = [255, 0, 0, 255]
        m.visual.face_colors = color
        # convert color visual to texture
        m.visual = m.visual.to_texture()
        # convert back to color
        m.visual = m.visual.to_color()
        assert g.np.allclose(m.visual.main_color, color)

    def test_pbr_export(self):
        # try loading a textured box
        m = next(iter(g.get_mesh('BoxTextured.glb').geometry.values()))
        # make sure material copy doesn't crash
        m.visual.copy()

        with g.TemporaryDirectory() as d:
            # exports by path allow files to be written
            path = g.os.path.join(d, 'box.obj')
            m.export(path)
            # try reloading
            r = g.trimesh.load(path)
            # make sure material survived
            assert r.visual.material.image.size == (256, 256)


if __name__ == '__main__':
    g.trimesh.util.attach_to_log()
    g.unittest.main()
